#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['parametrize_from_file']

package_data = \
{'': ['*']}

install_requires = \
['toml',
 'pyyaml',
 'nestedtext',
 'tidyexc',
 'pytest',
 'more_itertools',
 'contextlib2']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc', 'sphinx-toolbox'],
 'test': ['pytest', 'pytest-cov', 'coveralls', 'voluptuous', 'numpy']}

setup(name='parametrize_from_file',
      version='0.5.1',
      description='Parametrize test functions with values read from config files.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/parametrize_from_file',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.6',
     )
